/*
 * dyndns.c:
 * Submit a changed record into the dynamic DNS.
 *
 * This should be setuid to the dyndns user.
 *
 * Copyright (c) 2002 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: dyndns.c,v 1.4 2002/07/08 18:02:14 chris Exp $";

#include <gdbm.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define DYNDNS_DB_USER      "/var/dyndns/user"
#define DYNDNS_DB_HOST      "/var/dyndns/host"

/* usage:
 * Print usage. */
static void usage(FILE *fp) {
    fprintf(fp,
"Usage: dyndns <host> <numeric IP address>\n"
            );
}

static void dyndns_gdbm_abort(char *str) {
    fprintf(stderr, "dyndns: gdbm: %s\n", str);
    exit(1);
}

int main(int argc, char *argv[]) {
    GDBM_FILE db_user, db_host;
    struct in_addr addr;
    int i, ret = 1;
    datum key, val;

    if (argc != 3) {
        usage(stderr);
        exit(1);
    }

    if (!inet_aton(argv[2], &addr)) {
        fprintf(stderr, "dyndns: `%s' is not a valid numeric IP address\n", argv[2]);
        exit(1);
    }

    /* Open the two databases. We need to read from the user database and
     * perhaps write to the host database. Open the host database first, since
     * this effectively gives us a lock on the whole enterprise. */

    /* Host database. */
    for (i = 0; i < 3; ++i) {
        db_host = gdbm_open(DYNDNS_DB_HOST, 0, GDBM_WRITER | GDBM_SYNC, 0600, dyndns_gdbm_abort);
        if (db_host) break;
        else sleep(1 << i);
    }

    if (!db_host) {
        fprintf(stderr, "dyndns: %s: %s\n", DYNDNS_DB_HOST, gdbm_strerror(gdbm_errno));
        exit(1);
    }

     for (i = 0; i < 3; ++i) {
        db_user = gdbm_open(DYNDNS_DB_USER, 0, GDBM_READER, 0600, dyndns_gdbm_abort);
        if (db_user) break;
        else sleep(1 << i);
    }

    if (!db_user) {
        fprintf(stderr, "dyndns: %s: %s\n", DYNDNS_DB_USER, gdbm_strerror(gdbm_errno));
        gdbm_close(db_host);
        exit(1);
    }

    /* Check that this host is owned by the calling user. */
    key.dptr = argv[1];
    key.dsize = strlen(argv[1]) + 1;

    val = gdbm_fetch(db_user, key);
    if (!val.dptr)
        fprintf(stderr, "dyndns: %s: no such host\n", argv[1]);
    else if (val.dptr[val.dsize - 1] || strspn(val.dptr, "0123456789") != val.dsize - 1)
        fprintf(stderr, "dyndns: %s: bad database record\n", argv[1]);
    else if (getuid() != (uid_t)atoi(val.dptr))
        fprintf(stderr, "dyndns: %s: Permission denied\n", argv[1]);
    else {
        /* OK. Verify that the IP address has actually changed. */
        char *ad;
        ad = inet_ntoa(addr); /* canonical address */
        val = gdbm_fetch(db_host, key);
        if (!val.dptr || val.dsize != strlen(ad) + 1 || memcmp(val.dptr, ad, val.dsize) != 0) {
            val.dptr = ad;
            val.dsize = strlen(ad) + 1;
            gdbm_store(db_host, key, val, GDBM_REPLACE);
            ret = 0;

            /* Store also a record saying that an address has been changed.
             * Since underscores may not appear in domain names, call this
             * record __changed__. */
#define CHANGEDKEY      "__changed__";
            key.dptr = CHANGEDKEY;
            key.dsize = sizeof CHANGEDKEY;
            val.dptr = "1";
            val.dsize = 2;
            gdbm_store(db_host, key, val, GDBM_REPLACE);

            openlog("dyndnsd", LOG_PID, LOG_DAEMON);
            syslog(LOG_INFO, "%s IN A %s by uid %d", argv[1], ad, getuid());
            closelog();
        }
    }

    /* Done. */

finish:
    gdbm_close(db_host);
    gdbm_close(db_user);
   
    return ret;
}
